/*
 * File: interp.c
 * Author: Saumya Debray
 * Purpose: Define an interpreter for a very simple byte-code instruction set.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#include "interp.h"

/*********************************************************************
 *                                                                   *
 *                        REGISTERS AND MEMORY                       *
 *                                                                   *
 *********************************************************************/

int Regs[NREGS];
int Memory[NMEM];

extern Instr byteCodeArray[];
extern int nInstrs;

int OperandValue(Operand *op) {
  switch (op->type) {
  case REG:
    return Regs[op->value];
    
  case MEM:
    return Memory[op->value];
    
  case IMM:
    return op->value;
    
  /*
   * indirect operands: the field op->value is the register number to dereference.
   */
  case INDIR:
    return Memory[Regs[op->value]];

  default:
    fprintf(stderr, "ERROR [OperandValue]: Unknown operand type %d\n", op->type);
    exit(-1);
  }
  /* NOTREACHED */
}

/*
 * WriteBack(dest, val) -- write val into destination.
 */
void WriteBack(Operand *op, int val) {
  if (op->type == REG) {
    Regs[op->value] = val;
  }
  else if (op->type == MEM) {
    Memory[op->value] = val;
  }
  else {
    fprintf(stderr, "ERROR [WriteBack]: Illegal operand type %d\n", op->type);
    exit(-1);
  }
  /* NOTREACHED */
}

/*
 * interp(ip) -- interpret the program starting at the instruction
 * at index ip.
 */
void interp(int ip) {
  Instr *currInstr;
  int op1, op2, result;
  
  while (true) {
    currInstr =  &(byteCodeArray[ip]);
    switch (currInstr->opcode) {
    case MOV:
      WriteBack(&(currInstr->dest), OperandValue(&(currInstr->src1)));
      ip++;
      break;
    case ADD:
      op1 = OperandValue(&(currInstr->src1));
      op2 = OperandValue(&(currInstr->src2));
      result = op1 + op2;
      WriteBack(&(currInstr->dest), result);
      ip++;
      break;
    case SUB:
      op1 = OperandValue(&(currInstr->src1));
      op2 = OperandValue(&(currInstr->src2));
      result = op1 - op2;
      WriteBack(&(currInstr->dest), result);
      ip++;
      break;
    case MUL:
      op1 = OperandValue(&(currInstr->src1));
      op2 = OperandValue(&(currInstr->src2));
      result = op1 * op2;
      WriteBack(&(currInstr->dest), result);
      ip++;
      break;
    case DIV:
      op1 = OperandValue(&(currInstr->src1));
      op2 = OperandValue(&(currInstr->src2));
      result = op1 / op2;
      WriteBack(&(currInstr->dest), result);
      ip++;
      break;
    case IF_GE:
      op1 = OperandValue(&(currInstr->src1));
      op2 = OperandValue(&(currInstr->src2));
      if (op1 >= op2) {
	ip = OperandValue(&(currInstr->dest));
      }
      else {
	ip++;
      }
      break;
    case IF_GT:
      op1 = OperandValue(&(currInstr->src1));
      op2 = OperandValue(&(currInstr->src2));
      if (op1 > op2) {
	ip = OperandValue(&(currInstr->dest));
      }
      else {
	ip++;
      }
      break;
    case IF_LE:
      op1 = OperandValue(&(currInstr->src1));
      op2 = OperandValue(&(currInstr->src2));
      if (op1 <= op2) {
	ip = OperandValue(&(currInstr->dest));
      }
      else {
	ip++;
      }
      break;
    case IF_LT:
      op1 = OperandValue(&(currInstr->src1));
      op2 = OperandValue(&(currInstr->src2));
      if (op1 < op2) {
	ip = OperandValue(&(currInstr->dest));
      }
      else {
	ip++;
      }
      break;
    case IF_EQ:
      op1 = OperandValue(&(currInstr->src1));
      op2 = OperandValue(&(currInstr->src2));
      if (op1 == op2) {
	ip = OperandValue(&(currInstr->dest));
      }
      else {
	ip++;
      }
      break;
    case IF_NE:
      op1 = OperandValue(&(currInstr->src1));
      op2 = OperandValue(&(currInstr->src2));
      if (op1 != op2) {
	ip = OperandValue(&(currInstr->dest));
      }
      else {
	ip++;
      }
      break;
    case JMP:
      ip = OperandValue(&(currInstr->src1));
      break;
    case PRINT:
      op1 = OperandValue(&(currInstr->src1));
      printf("%d\n", op1);
      ip++;
      break;
    case HALT:
      return;
    }
  }
}

/*
 * init_mem() : initialize memory with the values to search.  Memory is used in
 * the following way:
 *
 *	Memory[0]: the key to search
 *	Memory[1]: number of values in the search array
 *	Memory[2]: lower bound array index to start the search (= 4)
 *	Memory[3]: upper bound array index to start the search
 *	Memory[4] ... Memory[???] : values to search
 */
void init_mem() {
  int search_key, nvals, i;
  
  if (scanf("%d", &search_key) != 1) {
    fprintf(stderr, "Could not read search key -- aborting\n");
    exit(1);
  }
  
  if (scanf("%d", &nvals) != 1) {
    fprintf(stderr, "Could not read search key -- aborting\n");
    exit(1);
  }
  /*
   * read in the array of values to search
   */
  for (i = 0; i < nvals; i++) {
    if (scanf("%d", &(Memory[4+i])) != 1) {
      fprintf(stderr, "Could not read array value for i = %d\n", i);
      exit(1);
    }
  }
  /*
   * initialize search parameters
   */  
  Memory[0] = search_key;
  Memory[1] = nvals;
  Memory[2] = 4;
  Memory[3] = 4+nvals-1;
}

int main() {
  init_mem();

  interp(0);

  return 0;
}
