/**
 * This file contains all of the code that is required exclusively to parse the DataOps trace format. 
 *  Anything that may be useful for another trace format reader I have tried to extract into ReaderUtils.c.
 *  Thus, this file almost exclusively consists of the code necessary to initialize/fetch/free the state 
 *  required to read the trace format and read data from the trace.
 **/

#include <assert.h>
#include "Buf.h"
#include "XedDisassembler.h"
#include "../shared/DataOpsDefs.h"
#include "ShadowMemory.h"
#include "ShadowRegisters.h"
#include "Verify.h"

typedef struct PredictionRecord_t {
    uint32_t predSrcId;
    uint32_t predFnId;
    uint64_t predAddr;
} PredictionRecord;

// trace-specific data
static char *strTable = NULL;
static uint16_t selections = 0;

// architecture-dependent data
static uint8_t addrSize;
static uint64_t (*parseAddr)(uint8_t *);
static uint8_t (*archLynxRegSize)(LynxReg);
LynxReg (*LynxRegFull) (LynxReg reg);

// Prediction data
static uint32_t predTID;
static PredictionRecord *predictions;
static uint32_t numSkips = 0;

// File data
Buf *traceBuf;
uint8_t *tracePos;
Buf *dataBuf;
uint8_t *dataPos;

// data-section data
uint64_t nextDataId;

/**
 * Function: parseDataOpsData
 * Description: Parses the next data section in the trace. It should be noted, that it does not check to see
 *  if the id of the next label matches the current id, so only call this function when the two match. It
 *  takes in the optional InsInfo parameter. If provided, it will check the state of the data it reads in
 *  to see if it agrees with our data. Either way, however, it reads in data and updates the Shadow Arch.
 *  appropriately.
 * Side Effects: The Shadow Architecture is updated with values from data and nextDataId is updated with
 *  the event id of the next label in the data section. If info is provided, the data will be checked which
 *  automatically outputs warnings to the user.
 * Outputs: N
 **/
void parseDataOpsData(InsInfo *info) {
    //try to get the first byte, if we can't then we are at the end of the data section
    dataPos = loadN(dataBuf, dataPos, 1);
    if(dataPos == NULL) {
        return;
    }

    initializeMarks(info);

    uint8_t size;
    uint8_t opType;

    //get buf
    while(*dataPos != OP_LABEL) {
        switch(opType = *dataPos) {
            case OP_REG: {
                dataPos += 1;
                dataPos = loadN(dataBuf, dataPos, 5);
                if(dataPos == NULL) {
                    throwError("Data Section ended in the middle of data");
                }

                uint32_t tid = *((uint32_t *) dataPos);
                dataPos += 4;

                LynxReg lReg = (LynxReg) *dataPos;
                dataPos += 1;

                size = archLynxRegSize(lReg);

                dataPos = loadN(dataBuf, dataPos, size);
                if(dataPos == NULL) {
                    throwError("Data Section ended in the middle of data");
                }

                setThreadVal(lReg, tid, dataPos);
                dataPos += size;
                break;
            } 
            case OP_DST_MEM: 
            case OP_MEM: {
                dataPos += 1;
                dataPos = loadN(dataBuf, dataPos, addrSize + 1);
                if(dataPos == NULL) {
                    throwError("Data Section ended in the middle of data");
                }

                size = *dataPos;
                dataPos += 1;

                uint64_t memAddr = parseAddr(dataPos);
                dataPos += addrSize;

                dataPos = loadN(dataBuf, dataPos, size);
                if(dataPos == NULL) {
                    throwError("Data Section ended in the middle of data");
                }

                store(memAddr, size, dataPos);
                dataPos += size;

                if(info != NULL && opType == OP_DST_MEM) {
                    findDstMem(memAddr, size, info);
                }
                break;
            } 
            case OP_DST_REG: {
                dataPos += 1;
                dataPos = loadN(dataBuf, dataPos, 5);
                if(dataPos == NULL) {
                    throwError("Data Section ended in the middle of data");
                }

                LynxReg dstReg = (LynxReg) *dataPos;
                LynxReg fullReg = LynxRegFull(dstReg);
                dataPos += 1;

                size = archLynxRegSize(fullReg);

                dataPos = loadN(dataBuf, dataPos, size);
                if(dataPos == NULL) {
                    throwError("Data Section ended in the middle of data");
                }

                setThreadVal(fullReg, predTID, dataPos);
                dataPos += size;

                if(info != NULL) {
                    findDstReg(dstReg, info);
                }
                break;
            } 
            case OP_TEST_REG: {
                dataPos += 1;
                dataPos = loadN(dataBuf, dataPos, 5);
                if(dataPos == NULL) {
                    throwError("Data Section ended in the middle of data");
                }

                LynxReg testReg = (LynxReg) *dataPos;
                LynxReg fullReg = LynxRegFull(testReg);
                dataPos += 1;

                size = archLynxRegSize(fullReg);

                dataPos = loadN(dataBuf, dataPos, size);
                if(dataPos == NULL) {
                    throwError("Data Section ended in the middle of data");
                }

                if(info != NULL) {
                    findSrcReg(testReg, info);

                    const uint8_t *regVal = getThreadVal(fullReg, predTID);

                    char msg[256];
                    sprintf(msg, "Register value %s mismatch: %s", LynxReg2Str(fullReg), info->mnemonic);

                    checkVals(dataPos, regVal, size, msg);
                }

                dataPos += size;
                break;
            } 
            case OP_TEST_MEM: {
                dataPos += 1;
                dataPos = loadN(dataBuf, dataPos, addrSize + 1);
                if(dataPos == NULL) {
                    throwError("Data Section ended in the middle of data");
                }

                size = *dataPos;
                dataPos += 1;

                uint64_t memAddr = parseAddr(dataPos);
                dataPos += addrSize;

                dataPos = loadN(dataBuf, dataPos, size);
                if(dataPos == NULL) {
                    throwError("Data Section ended in the middle of data");
                }
 
                if(info != NULL) {
                    findSrcMem(memAddr, size, info);

                    uint8_t memVal[64];
                    load(memAddr, size, memVal);

                    char msg[256];
                    sprintf(msg, "Memory value at %llx mismatch: %s", (unsigned long long) memAddr, info->mnemonic);
                    checkVals(dataPos, memVal, size, msg);
                }

                dataPos += size;
                break;
            } 
            default:
                dataPos += 1;
                throwError("Invalid data operand type found");
        }

        //try to get the next label, if we can't that's fine because we are at the beginning of new data
        dataPos = loadN(dataBuf, dataPos, 1);
        if(dataPos == NULL) {
            return;
        }
    }

    dataPos += 1;

    //we got a label, but no event id
    dataPos = loadN(dataBuf, dataPos, 8);
    if(dataPos == NULL) {
        throwError("Data Section ended in the middle of data");
    } 

    nextDataId = *((uint64_t *) dataPos);
    dataPos += 8;

    if(info != NULL) {
        checkMarks(info, selections);
    }
}

/**
 * Function: initDataOps
 * Description: Initializes the reader for the DataOps format. This requires setting up any architecture
 *  specific requirements, reading any format-specific headers, and initializing any state. One important
 *  thing to note is that this will call the data parsing function so that the initial state gets set.
 * Side Effects: Quite a bit gets initialized
 * Output: None
 **/
void initDataOps (FileHeader *fileHeader, SectionEntry * sectionTable, FILE *traceFile) {
    uint16_t i;

    TraceHeader traceHeader;
    InfoSelHeader infoSelHeader;

    uint8_t foundTrace = 0;
    uint8_t foundData = 0;
    uint8_t foundStrTable = 0;
    uint8_t foundInfoSelHeader = 0;
    uint8_t foundTraceHeader = 0;

    switch(fileHeader->archType) {
        case X86_ARCH:
            parseAddr = parseAddr32;
            addrSize = 4;
            archLynxRegSize = LynxRegSize32;
            LynxRegFull = LynxReg2FullLynxIA32Reg;
            break;
        case X86_64_ARCH:
            parseAddr = parseAddr64;
            addrSize = 8;
            archLynxRegSize = LynxRegSize;
            LynxRegFull = LynxReg2FullLynxIA32EReg;
            break;
        default:
            throwError("Unknown Architecture Type");
    }

    for(i = 0; i < fileHeader->sectionNumEntry; i++) {
        switch(sectionTable[i].type) {
            case TRACE_SECTION: //this section was read in Trd.c apparently
                foundTrace = 1;
                traceBuf = readTraceEntry(traceFile, sectionTable + i);
                tracePos = traceBuf->buf + 1;

                if(traceBuf == NULL) {
                    throwError("Unable to Read Trace Section");
                }
                break;
            case DATA_SECTION: // this section was read in Trd.c apparently
                foundData = 1;
                dataBuf = readDataEntry(traceFile, sectionTable + i);
                dataPos = dataBuf->buf;

                if(dataBuf == NULL) {
                    throwError("Unable to Read Data Section");
                }
                break;
            case TRACE_HEADER_SECTION:
                foundTraceHeader = 1;

                if(readTraceHeaderEntry(traceFile, sectionTable + i, &traceHeader) != 1) {
                    throwError("Unable to Read Trace Header");
                }
                break;
            case STR_TABLE:
                foundStrTable = 1;
                strTable = readStrTableEntry(traceFile, sectionTable + i);

                if(strTable == NULL) {
                    throwError("Unable to Read String Table");
                }
                break;
            case INFO_SEL_HEADER:
                foundInfoSelHeader = 1;

                if(readInfoSelHeaderEntry(traceFile, sectionTable + i, &infoSelHeader) != 1) {
                    throwError("Unable to Read Info Sel Header");
                }

                selections = infoSelHeader.selections;
                break;
        }
    }

    if(!foundTrace) {
        throwError("Unable to Locate Trace Section in File");
    }
    if(!foundData) {
        throwError("Unable to Locate Data Section in File");
    }
    if(!foundTraceHeader) {
        throwError("Unable to Locate Trace Header in File");
    }
    if(!foundStrTable) {
        throwError("Unable to Locate String Table in File");
    }
    if(!foundInfoSelHeader) {
        throwError("Unable to Locate Information Selection Header in File");
    }

    initShadowRegisters (traceHeader.numThreads);
    initShadowMemory ();

    predictions = malloc(sizeof(PredictionRecord) * traceHeader.numThreads);

    predTID = 0;
    nextDataId = 0;
    for(i = 0; i < traceHeader.numThreads; i++) {
        predictions[i].predSrcId = 0;
        predictions[i].predFnId = 0;
        predictions[i].predAddr = 0;
    }

    parseDataOpsData(NULL);
}

/**
 * Function: freeDataOps
 * Description: Frees state used by the DataOps format. One thing to note is that it currently does not
 *  free the string table so that if instructions are being kept around, the user can still get the name
 *  of libraries and functions used by the instruction.
 * Output: None
 **/
void freeDataOps() {
    freeBuf(traceBuf);
    freeBuf(dataBuf);
    free(predictions);
    //free(strTable);
}

/**
 * Function: genInsFromPred
 * Description: Generates an instruction from predicted values. Note that the predicted value for binary is
 *  the value stored at that address in shadow memory.
 * Side Effects: The input ReaderIns is populated
 * Output: None
 **/
static void genInsFromPred(ReaderIns *ins) {
    ins->binSize = 0;
    ins->tid = predTID;

    ins->srcId = predictions[ins->tid].predSrcId;

    ins->fnId = predictions[ins->tid].predFnId;

    ins->addr = predictions[ins->tid].predAddr;

    if(selections & (1 << SEL_BIN)) {
        xed_decoded_inst_t xedIns;
        load(ins->addr, 15, ins->binary);

        ins->binSize = decodeIns(&xedIns, ins->binary, 15);

        predictions[ins->tid].predAddr = ins->addr + ins->binSize;
        setThreadVal(LYNX_RIP, ins->tid, (uint8_t *)(&predictions[ins->tid].predAddr));
    }
}

/**
 * Function: parseIns
 * Description: Parses an instruction from the trace. Currently it has to save the position of quite a few
 *  data elements because in many cases the TID is required, which is among the last items to be parsed.
 *  Note that it will read the instruction form pos.
 * Side Effects: The input ReaderIns gets populated. The predictions are updated.
 * Output: The position in the buffer immediately after the instruction
 **/
static uint8_t *parseIns(ReaderIns *ins, uint8_t *pos) {
    uint8_t *startingPos = pos;
    uint32_t *srcIdPos = NULL;
    uint32_t *fnIdPos = NULL;
    uint8_t *addrPos = NULL;

    ins->binSize = 0;

    if((*pos & 0x7) != EVENT_INS) {
        throwError("Exception found was not an instruction");
    }
    uint8_t predResults = *pos & (~0x7);
    pos += 1;

    //get source
    if(!((predResults >> PRED_SRCID) & PRED_CORRECT)) {
        srcIdPos = (uint32_t *) pos;
        pos += 4;
    }

    if(!((predResults >> PRED_FUNCID) & PRED_CORRECT)) {
        fnIdPos = (uint32_t *) pos;
        pos += 4;
    }

    if(!((predResults >> PRED_ADDR) & PRED_CORRECT)) {
        addrPos = pos;
        pos += addrSize;
    }

    if(!((predResults >> PRED_BIN) & PRED_CORRECT)) {
        xed_decoded_inst_t xedIns;
        ins->binSize = decodeIns(&xedIns, pos, 15);
        
        memcpy(ins->binary, pos, ins->binSize);

        if(ins->binSize == 0) {
            throwError("Unable to parse binary from trace");
        }

        pos += ins->binSize;
    }

    if((predResults >> PRED_TID) & PRED_CORRECT) {
        ins->tid = predTID;
    }
    else {
        ins->tid = *((uint32_t *) pos);
        pos += 4;
        predTID = ins->tid;
    }


    if(srcIdPos == NULL) {
        ins->srcId = predictions[ins->tid].predSrcId;
    }
    else {
        ins->srcId = *srcIdPos;
        predictions[ins->tid].predSrcId = *srcIdPos;
    }

    if((predResults >> PRED_FUNCID) & PRED_CORRECT) {
        ins->fnId = predictions[ins->tid].predFnId;
    }
    else {
        ins->fnId = *fnIdPos;
        predictions[ins->tid].predFnId = *fnIdPos;
    }

    if((predResults >> PRED_ADDR) & PRED_CORRECT) {
        ins->addr = predictions[ins->tid].predAddr;
    }
    else {
        ins->addr = parseAddr(addrPos);
    }

    if(ins->binSize != 0) {
        store(ins->addr, ins->binSize, ins->binary);
    }
    else if(selections & (1 << SEL_BIN)) {
        xed_decoded_inst_t xedIns;
        load(ins->addr, 15, ins->binary);

        ins->binSize = decodeIns(&xedIns, ins->binary, 15);
        if(ins->binSize == 0) {
            throwError("Unable to read instruction binary from shadow memory");
        }
    }

    predictions[ins->tid].predAddr = ins->addr + ins->binSize;
    setThreadVal(LYNX_RIP, ins->tid, (uint8_t *)(&predictions[ins->tid].predAddr));

    if((pos - startingPos + 2) != *((uint16_t *) pos)) {
        throwError("Event size does not match expected size");
    }

    pos += 2;

    return pos;
}


/**
 * Function: parseException
 * Description: Parses an exception from the input buffer and populates exception.
 * Side Effects: The ReaderException is populated
 * Output: The position in the buffer just after the exception
 **/
static uint8_t *parseException(ReaderException *exception, uint8_t *buf) {
    uint8_t *startingPos = buf;
    if(*buf != EVENT_EXCEPT) {
        throwError("Event found was not an exception");
    }
    buf += 1;

    exception->type = *((uint8_t *) buf);
    buf += 1;

    if(exception->type != LINUX_SIGNAL && exception->type != WINDOWS_EXCEPTION) {
        throwError("Unknown Exception Type");
    }

    exception->code = *((uint32_t *) buf);
    buf += 4;

    exception->tid = *((uint32_t *) buf);
    buf += 4;

    exception->addr = parseAddr(buf);
    buf += addrSize;

    if((buf - startingPos + 2) != *((uint16_t *) buf)) {
        throwError("Event size does not matcch expected event size");
    }

    buf += 2;

    return buf;
}

/**
 * Function: readDataOpsEvent
 * Description: Performs all of the steps necessary to read the next instruction with the id curId. If
 *  the id of the next instruction matches the next label in the data section, data will be read first.
 *  Note, this is the only thing curInfo is needed for, and is allowed to be NULL since it is optional to
 *  provide curInfo to parseDataOpsData. After reading the data, it will read the next event or instruction
 *  from the trace and populate nextEvent.
 * Side Effects: nextEvent is populated
 * Output: 1 if an instruction was parsed, 0 otherwise
 **/
uint32_t readDataOpsEvent(ReaderEvent *nextEvent, uint64_t curId, InsInfo *curInfo) {
    if((curId - 1) == nextDataId) {
        parseDataOpsData(curInfo);
    }

    if(numSkips == 0) {
        uint8_t *tmpPos;
        if((tmpPos = loadN(traceBuf, tracePos, 38)) == NULL) {
            uint64_t remaining = getBytesRemaining(traceBuf, tracePos);
            if(remaining == 0) {
                return 0;
            }
            
            tmpPos = loadN(traceBuf, tracePos, remaining);
        }

        tracePos = tmpPos;

        switch(*(tracePos) & 0x7) {
            case EVENT_INS:
                nextEvent->type = INS_EVENT;
                tracePos = parseIns(&nextEvent->ins, tracePos);
                break;
            case EVENT_EXCEPT:
                nextEvent->type = EXCEPTION_EVENT;
                tracePos = parseException(&nextEvent->exception, tracePos);
                break;
            default:
                throwError("Unknown Event Type");
        }

        numSkips = *tracePos;
    }
    else {
        nextEvent->type = INS_EVENT;
        genInsFromPred(&nextEvent->ins);
        numSkips--;
    }

    //only pass numSkips byte once we have actually finished doing the correct number of skips. If
    //we don't, the reader could mark that we have finished the trace prematurely for traces with
    //skips at the end
    if(numSkips == 0) {
        tracePos = loadN(traceBuf, tracePos, 1);
        tracePos += 1;
    }

    return 1;
}

/**
 * Function: fetchDataOpsStr
 * Description: Returns the string with the given id from the string table.
 * Output: The sting with the matching id.
 **/
const char *fetchDataOpsStr(uint32_t id) {
    return strTable + id;
}
