/**
 * This file contains the implementation of all the functions in the Reader's public API. Additionally, it 
 *  includes all of the state necessary to support the public API and any helper functions that are only
 *  needed by the public API (such as reading the sections of the trace required by any formats). Note,
 *  descriptions for the public API are not included here since I didn't want to worry about one set of 
 *  descriptions getting outdated if a function changes.
 **/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <inttypes.h>
#include <string.h>
#include "ReaderUtils.h"
#include "Reader.h"
#include "../shared/header.h"
#include "ShadowMemory.h"
#include "ShadowRegisters.h"
#include "../shared/XedLynxReg.h"
#include "DataOpsReader.h"
#include "XedDisassembler.h"

//reader state
FileHeader fileHeader;
FILE *trace = NULL;
uint8_t debug;
uint64_t curId;

//format dependent data
void (*initFormat) (FileHeader *, SectionEntry *, FILE *traceFile);
uint32_t (*readEvent) (ReaderEvent *, uint64_t, InsInfo *);
void (*freeFormat) ();
const char *(*fetchStr) (uint32_t);

/**
 * readHeader(FILE *trace, fileHeader *header) - readHeader reads the fileHeader from the given file.
 * The head is stored into the global variable fileHeader. The value returned represents the number of 
 * fileheaders read (1 for success, 0 for failure).
 **/
uint32_t readHeader (FILE * trace, FileHeader * header) {
    uint32_t retVal = 0;
    if (fseek (trace, 0, SEEK_SET)) {
	    throwError("Couldn't seek to file header location");;
    }

    retVal = fread(header, sizeof (FileHeader), 1, trace);

    return retVal;
}

/**
 * readSectionTable(FILE *trace, sectionEntry **sectionTable) -- readSectionTable reads the section table from
 * the given file.  Assumes that the fileHeader fileHeader has been read in already.  The value returned 
 * represents the number of section entries read.  Note that an array is dynamically allocated at the 
 * address pointed to by sectionTable. This must be freed by the caller. 
 **/
uint32_t readSectionTable (FILE * trace, FileHeader *header, int numEntries, SectionEntry *table) {
    if (fseek (trace, header->sectionTableOff, SEEK_SET)) {
	    throwError("Couldn't seek to section table location");
    }

    uint32_t readAmt = fread (table, header->sectionEntrySize, numEntries, trace);

    return readAmt;
}


const char *fetchSrcName(ReaderIns *ins) {
    return fetchStr(ins->srcId);
}

const char *fetchFnName(ReaderIns *ins) {
    return fetchStr(ins->fnId);
}

uint64_t getCurEid() {
    return curId;
}

uint32_t nextEvent (ReaderEvent *event) {
    return nextEventWithCheck(event, NULL);
}

uint32_t nextEventWithCheck(ReaderEvent *event, InsInfo *info) {
    curId++;
    if(event == NULL) {
        throwError("Cannot pass a NULL event to nextReaderEvent");
    }

    return readEvent(event, curId, info);
}

void initReader (char *fileName, uint32_t debugFlag) {
    //we set curId to -1 since nextEvent will increment
    curId = -1;

    //make sure to opoen as a binary file
    trace = fopen (fileName, "rb");

    if (trace == NULL) {
    	throwError("Unable to open specified file");
    }

    debug = debugFlag;

    if (readHeader(trace, &fileHeader) != 1) {
	    throwError("Unable to read trace header");
    }

    //perform architecture-dependent setup
    switch (fileHeader.archType) {
	    case X86_ARCH:
            setupDisassembler_x86();
	        break;
	    case X86_64_ARCH:
            setupDisassembler_x64();
	        break;
	    default:
	        throwError("Unknown architecture");
    }

    //perform setup based on trace type
    switch (fileHeader.traceType) {
        case DATA_OPS_TRACE:
            initFormat = initDataOps;
            readEvent = readDataOpsEvent;
            freeFormat = freeDataOps;
            fetchStr = fetchDataOpsStr;
            break;
	    default:
	        throwError("Unsupported trace type");
    }

    SectionEntry *sectionTable = malloc(fileHeader.sectionNumEntry * fileHeader.sectionEntrySize);

    if(sectionTable == NULL) {
        throwError("Could not allocate memory for section table");
    }

    if(readSectionTable(trace, &fileHeader, fileHeader.sectionNumEntry, sectionTable) != fileHeader.sectionNumEntry) {
        free(sectionTable);
        throwError("Could not read section table");
    }

    //allow the selected format to do any necessary setup
    initFormat (&fileHeader, sectionTable, trace);

    //we no longer need the section table
    free (sectionTable);
}

void initInsInfo(InsInfo *info) {
    int i;

    //srcOps, dstOps and readWriteOps are all partially statically defined linked lists, so link them 
    // appropriately
    for(i = 0; i < 2; i++) {
        info->srcOps[i].type = NONE_OP;
        info->srcOps[i].next = &info->srcOps[i + 1];
    }

    info->srcOpCnt = 0;
    info->srcOps[2].type = NONE_OP;
    info->srcOps[2].next = NULL;

    info->dstOpCnt = 0;
    info->dstOps[0].type = NONE_OP;
    info->dstOps[0].next = NULL;

    info->readWriteOpCnt = 0;
    info->readWriteOps[0].type = NONE_OP;
    info->readWriteOps[0].next = NULL;
}

int fetchInsInfo(ReaderIns *ins, InsInfo *info) {
    //reset counts, but note the linked list does NOT get modified, that way we can take advantage of 
    // past allocations
    info->srcOpCnt = 0;
    info->dstOpCnt = 0;
    info->readWriteOpCnt = 0;

    xed_decoded_inst_t xedIns;
    decodeIns(&xedIns, ins->binary, 15);

    info->insClass = getInsClass(&xedIns);
    getInsMnemonic(&xedIns, info->mnemonic, 128);
    getInsOps(&xedIns, ins->tid, info->srcOps, &info->srcOpCnt, info->dstOps, &info->dstOpCnt, info->readWriteOps, &info->readWriteOpCnt);

    return 0;
}

void closeReader (void) {
    if (trace != NULL) {
	    fclose (trace);
    }

    // cannot access shadow architecture after closing the reader
    freeShadowMemory ();
    freeShadowRegisters ();
    freeFormat();
}

void getMemoryVal (uint64_t addr, uint32_t size, uint8_t * buf) {
    load (addr, size, buf);
}

const uint8_t *getRegisterVal (LynxReg reg, uint32_t thread) {
    return getThreadVal (reg, thread);
}
