/*
* File: ShadowMemory.c
* Authors: Jon Stephens and David Raicher
* Description: This provides a Shadow Memory backend for the trace reader.  It does this implementing a quad map structure to provide byte level
* addressing for memory reads and writes.  The main interfaces for interacting with the Shadow Memory are the store and load functions.  If you are
* doing development on this code please note that there are 1/2/4/8 bit versions of store/load internally to allow for faster reads/writes.
*/

#include "ShadowMemory.h"
#include "ReaderUtils.h"

#define SM QM***
#define TM QM**
#define ADDRINT uint64_t
#define LVL_WIDTH 16
#define LVL_SIZE (1 << LVL_WIDTH)
#define PM_SHAMT (LVL_WIDTH * 3)
#define SM_SHAMT (LVL_WIDTH * 2)
#define TM_SHAMT LVL_WIDTH
#define ADDR_MASK (LVL_SIZE - 1)

typedef struct QM_t
{
    uint8_t val[LVL_SIZE];
} QM;

QM ***PM[LVL_SIZE];		// default secondary map
QM ***DSM;			// 256
QM **DTM;
QM *DQM;

/*
* initShadowMemory() -- Initializes the Shadow Memory.  This must be called before any other Shadow Memory functions are called.
* freeShadowMemory should also be called before any subsequent calls to this function to prevent memory leakage.
*/

void
initShadowMemory (void)
{
    DSM = malloc(sizeof (QM **[LVL_SIZE]));
    DTM = malloc(sizeof (QM *[LVL_SIZE]));
    DQM = malloc(sizeof (QM));

    if(DSM == NULL || DTM == NULL || DQM == NULL) {
        throwError("Unable to allocate space for default maps in shadow memory");
    }
    uint32_t i;
    for (i = 0; i < LVL_SIZE; i++) {
	PM[i] = DSM;
	DSM[i] = DTM;
	DTM[i] = DQM;
    }
}

static inline bool
is_DTM (TM curTM)
{
    return DTM == curTM;
}

static inline bool
is_DSM (SM curSM)
{
    return DSM == curSM;
}

static inline bool
is_DQM (QM * curQM)
{
    return DQM == curQM;
}

/*
* reset() -- Resets the state of the shadow memory to be completely clean (that is, just like after a call to initShadowMemory()). 
*/

void
reset (void)
{
    uint32_t i, j, k;
    for (i = 0; i < LVL_SIZE; i++) {
	SM curSM = PM[i];
	if (DSM == NULL || !is_DSM (curSM)) {
	    for (j = 0; j < LVL_SIZE; j++) {
		TM curTM = curSM[j];
		if (!is_DTM (curTM)) {
		    for (k = 0; k < LVL_SIZE; k++) {
			QM *curQM = curTM[k];
			if (!is_DQM (curQM)) {
			    free (curQM);
			    curTM[k] = DQM;
			}
		    }

		    free (curTM);
		    curSM[j] = DTM;
		}
	    }
	    if (DSM != NULL) {
		free (curSM);
		PM[i] = DSM;
	    }
	}
    }
}

/*
* freeShadowMemory() -- Frees all memory allocated by the Shadow Memory.
*/

void
freeShadowMemory (void)
{
    reset ();
    free (DSM);
    free (DTM);
    free (DQM);
}

static inline SM
getSM (ADDRINT addr)
{
    return PM[addr >> PM_SHAMT];
}

static inline TM
getTM (ADDRINT addr, SM curSM)
{
    return curSM[(addr >> SM_SHAMT) & ADDR_MASK];
}

static inline QM *
getQM (ADDRINT addr, TM curTM)
{
    return curTM[(addr >> TM_SHAMT) & ADDR_MASK];
}

static inline SM *
getSMPtr (ADDRINT addr)
{
    return &PM[addr >> PM_SHAMT];
}

static inline TM *
getTMPtr (ADDRINT addr, SM curSM)
{
    return &curSM[(addr >> SM_SHAMT) & ADDR_MASK];
}

static inline QM **
getQMPtr (ADDRINT addr, TM curTM)
{
    return &curTM[(addr >> TM_SHAMT) & ADDR_MASK];
}

/*
* copyForWritingTM(ADDRINT addr, TM *srcTM) -- Replaces TM in primary map, and returns
* created QM.
*/

static inline QM *
copyForWritingTM (ADDRINT addr, TM * srcTM)
{
    TM newTM = malloc (sizeof (QM *[LVL_SIZE]));

    if(newTM == NULL) {
        throwError("Unable to allocate space for new TM");
    }

    memcpy (newTM, *srcTM, sizeof (QM *[LVL_SIZE]));

    QM *newQM = malloc(sizeof (QM));

    if(newQM == NULL) {
        throwError("Unable to allocate space for new QM");
    }

    QM **curQM = getQMPtr (addr, newTM);
    memcpy (newQM, *curQM, sizeof (QM));

    *srcTM = newTM;
    *curQM = newQM;

    return newQM;
}

/*
* copyForWritingSM(ADDRINT addr, TM *srcTM) -- Replaces SM in primary map, and returns
* created QM.
*/

static inline QM *
copyForWritingSM (ADDRINT addr, SM * srcSM)
{
    SM newSM = malloc(sizeof (QM **[LVL_SIZE]));

    if(newSM == NULL) {
        throwError("Unable to allocate space for new SM");
    }

    memcpy (newSM, *srcSM, sizeof (QM **[LVL_SIZE]));

    TM newTM = malloc(sizeof (QM *[LVL_SIZE]));

    if(newTM == NULL) {
        throwError("Unable to allocate space for new TM");
    }

    TM *curTM = getTMPtr (addr, newSM);
    memcpy (newTM, *curTM, sizeof (QM *[LVL_SIZE]));

    QM *newQM = malloc(sizeof (QM));

    if(newQM == NULL) {
        throwError("Unable to allocate space for new QM");
    }

    QM **curQM = getQMPtr (addr, newTM);
    memcpy (newQM, *curQM, sizeof (QM));

    *srcSM = newSM;
    *curTM = newTM;
    *curQM = newQM;

    return newQM;
}

/*
* copyForWritingSM(ADDRINT addr, TM *srcTM) -- Replaces QM in primary map, and returns
* created QM.
*/

static inline QM *
copyForWritingQM (ADDRINT addr, QM ** srcQM)
{
    QM *newQM = malloc(sizeof (QM));

    if(newQM == NULL) {
        throwError("Unable to allocate space for new QM");
    }

    memcpy (newQM, *srcQM, sizeof (QM));

    *srcQM = newQM;

    return newQM;
}

/*
* getQMForReading(ADDRINT addr) -- Returns the QM at the specified address for reading.
*/

static inline QM *
getQMForReading (ADDRINT addr)
{
    return PM[addr >> PM_SHAMT][(addr >> SM_SHAMT) &
				ADDR_MASK][(addr >> TM_SHAMT) & ADDR_MASK];
}

/*
* getQMForWriting(ADDRINT addr) -- Returns the QM for the specified address.
*/

static inline QM *
getQMForWriting (ADDRINT addr)
{
    SM *curSM = getSMPtr (addr);
    if (is_DSM (*curSM)) {
	return copyForWritingSM (addr, curSM);
    }

    TM *curTM = getTMPtr (addr, *curSM);
    if (is_DTM (*curTM)) {
	return copyForWritingTM (addr, curTM);
    }

    QM **curQM = getQMPtr (addr, *curTM);
    if (is_DQM (*curQM)) {
	return copyForWritingQM (addr, curQM);
    }

    return *curQM;
}

/*
* isDefault() -- Checks to see which chunks of memory have been allocated (and hence are no longer the default map values).
*/

void
isDefault (void)
{
    uint32_t i;
    for (i = 0; i < LVL_SIZE; i++) {
	if (DSM != NULL && !is_DSM (PM[i])) {
	    printf ("Not DSM at %d\n", i);
	}
	if (DTM != NULL && !is_DTM (DSM[i])) {
	    printf ("Not DTM at %d\n", i);
	}
	if (DQM != NULL && !is_DQM (DTM[i])) {
	    printf ("Not DQM at %d\n", i);
	}
    }
}

/*
* chkDefaultVals() -- Checks to ensure the default map values are consistent.
*/

void
chkDefaultVals (void)
{
    uint32_t i;
    for (i = 0; i < LVL_SIZE; i++) {
	if (DSM != NULL && DSM[i] != DTM) {
	    printf ("Bad DSM at %d\n", i);
	}
	if (DTM != NULL && DTM[i] != DQM) {
	    printf ("Bad DTM at %d\n", i);
	}
	if (i < LVL_SIZE && DQM->val[i] != 0) {
	    printf ("Bad DQM at %p, val: %x\n", DQM, DQM->val[i]);
	}
    }
}

/*
* isNotAligned(ADDRINT addr, uint32_t alignment) -- Checks to see if the size of a given read/write is aligned with
* the QM size.  Easy check to see if a write/read may be split between seperate QMs.
*/

static inline bool
isNotAligned (ADDRINT addr, uint32_t alignment)
{
    return addr & (alignment - 1);
}

/*
* hasSpace(ADDRINT addr, unit32_t sizeInBytes) -- Determines if there is space on the given QM starting at addr for the
* given write.  Can be used to ensure there isn't a write off the side of the array. Returns 0 if there is not space.  1 otherwise.
*/

static inline bool
hasSpace (ADDRINT addr, uint32_t sizeInBytes)
{
    return ((addr & ADDR_MASK) + sizeInBytes) < ADDR_MASK;
}

/*
* store8(ADDRINT addr, uint8_t newVal) -- Stores newVal (a byte) at the given address.
*/

static inline void
store8 (ADDRINT addr, uint8_t newVal)
{
    QM *curQM = getQMForWriting (addr);
    uint32_t i = (addr & ADDR_MASK);
    curQM->val[i] = newVal;
}

/*
* storeSlow(ADDRINT addr, uint8_t *val, uint32_t sizeInBytes) -- Stores the bytes in the buffer val starting at addr one by one.  Obviously
* the slowest way to store large values.  Assume val is at least sizeInBytes long.
*/

void
storeSlow (ADDRINT addr, uint8_t * val, uint32_t sizeInBytes)
{
    uint32_t i;
    for (i = 0; i < sizeInBytes; i++) {
	store8 (addr + i, *(val + i));
    }
}

/*
* store16/store32/store64(ADDRINT, addr, uint8_t *val) -- Quickly stores 2/4/8 bytes from val into shadow memory starting at address addr.  Assumes
* That val is long enough for the write.
*/

static inline void
store16 (ADDRINT addr, uint8_t * val)
{
    if (isNotAligned (addr, 2)) {
	storeSlow (addr, val, 2);
	return;
    }
    QM *curQM = getQMForWriting (addr);
    uint32_t i = addr & ADDR_MASK;
    uint8_t *srcBuf = curQM->val + i;
    *((uint16_t *) srcBuf) = *((uint16_t *) val);
}

static inline void
store32 (ADDRINT addr, uint8_t * val)
{
    if (isNotAligned (addr, 4)) {
	storeSlow (addr, val, 4);
	return;
    }
    QM *curQM = getQMForWriting (addr);
    uint32_t i = addr & ADDR_MASK;
    uint8_t *srcBuf = curQM->val + i;
    *((uint32_t *) srcBuf) = *((uint32_t *) val);
}

static inline void
store64 (ADDRINT addr, uint8_t * val)
{
    if (isNotAligned (addr, 8)) {
	storeSlow (addr, val, 8);
	return;
    }
    QM *curQM = getQMForWriting (addr);
    uint32_t i = addr & ADDR_MASK;
    uint8_t *srcBuf = curQM->val + i;
    *((uint64_t *) srcBuf) = *((uint64_t *) val);
}

/*
* store(uint64_t addr, uint32_t sizeInBytes, uint8_t *val) --Stores the series of bytes (up to sizeInBytes) stored in val into the shadow 
* memory at address addr.  Assumes val is at least sizeInBytes long.
*/

void
store (ADDRINT addr, uint32_t sizeInBytes, uint8_t * val)
{
    if (sizeInBytes <= 8) {
	if (sizeInBytes == 4) {
	    store32 (addr, val);
	    return;
	}
	else if (sizeInBytes == 8) {
	    store64 (addr, val);
	    return;
	}
	else if (sizeInBytes == 2) {
	    store16 (addr, val);
	    return;
	}
	else if (sizeInBytes == 1) {
	    store8 (addr, *val);
	    return;
	}
    }

    while (sizeInBytes != 0 && isNotAligned (addr, 8)) {
	store8 (addr, *val);
	addr++;
	val++;
	sizeInBytes--;
    }

    while (sizeInBytes >= 8) {
	store64 (addr, val);
	sizeInBytes -= 8;
	val += 8;
	addr += 8;
    }

    while (sizeInBytes > 0) {
	store8 (addr, *val);
	addr++;
	val++;
	sizeInBytes--;
    }
}

/*
* load8(ADDRINT addr) -- Returns the value currently stored at addr (1 byte).
*/

static inline uint8_t
load8 (ADDRINT addr)
{
    QM *curQM = getQMForReading (addr);
    uint8_t retVal = curQM->val[addr & ADDR_MASK];
    return retVal;
}

/*
* loadSlow(ADDRINT addr, uint32_t sizeInBytes, uint8_t *buffer) -- Loads sizeInBytes bytes into buffer starting at addr.  This is done
* bytewise and hence is the slowest possible way.  Assumes buffer is at least sizeInBytes long.
*/

void
loadSlow (ADDRINT addr, uint32_t sizeInBytes, uint8_t * buffer)
{
    uint32_t i;
    for (i = 0; i < sizeInBytes; i++) {
	*(buffer + i) = load8 (addr + i);
    }
}

/*
* load16/load32/load64(ADDRINT, addr, uint8_t *buffer) -- Quickly loads 2/4/8 bytes from shadow memory into buffer starting at address addr.  Assumes
* That buffer is sizeInBytes long.
*/

static inline void
load16 (ADDRINT addr, uint8_t * buffer)
{
    if (isNotAligned (addr, 2)) {
	loadSlow (addr, 2, buffer);
    }
    QM *curQM = getQMForReading (addr);
    uint32_t i = addr & ADDR_MASK;
    uint8_t *destBuf = curQM->val + i;
    *((uint16_t *) buffer) = *((uint16_t *) destBuf);
}

static inline void
load32 (ADDRINT addr, uint8_t * buffer)
{
    if (isNotAligned (addr, 4)) {
	loadSlow (addr, 4, buffer);
    }
    QM *curQM = getQMForReading (addr);
    uint32_t i = addr & ADDR_MASK;
    uint8_t *destBuf = curQM->val + i;
    *((uint32_t *) buffer) = *((uint32_t *) destBuf);
}

static inline void
load64 (ADDRINT addr, uint8_t * buffer)
{
    if (isNotAligned (addr, 8)) {
	loadSlow (addr, 8, buffer);
    }
    QM *curQM = getQMForReading (addr);
    uint32_t i = addr & ADDR_MASK;
    uint8_t *destBuf = curQM->val + i;
    *((uint64_t *) buffer) = *((uint64_t *) destBuf);
}

/* 
* load(uint64_t addr, uint32_t size, uint8_t *buf) -- Retrieves size bytes from the address at addr in shadow memory 
* and stores them into the buffer at buf. Assumes buf is at least size long.
*/

void
load (ADDRINT addr, uint32_t sizeInBytes, uint8_t * buffer)
{
    if (sizeInBytes <= 8) {
	if (sizeInBytes == 4) {
	    load32 (addr, buffer);
	    return;
	}
	else if (sizeInBytes == 8) {
	    load64 (addr, buffer);
	    return;
	}
	else if (sizeInBytes == 2) {
	    load16 (addr, buffer);
	    return;
	}
	else if (sizeInBytes == 1) {
	    buffer[0] = load8 (addr);
	    return;
	}
    }

    uint32_t i = 0;

    while (sizeInBytes != 0 && isNotAligned (addr, 8)) {
	buffer[i++] = load8 (addr++);
	sizeInBytes--;
    }

    while (sizeInBytes >= 8) {
	load64 (addr, (buffer + i));
	i += 8;
	addr += 8;
	sizeInBytes -= 8;
    }

    while (sizeInBytes > 0) {
	buffer[i++] = load8 (addr++);
	sizeInBytes--;
    }
}
