/*
* File: ShadowMemory.h
* Author: Jon Stephens and David Raicher
* Description: This provides a Shadow Memory backend for the trace reader.  It does this implementing a quad map structure to provide byte level
* addressing for memory reads and writes.  The main interfaces for interacting with the Shadow Memory are the store and load functions.
*/

#ifndef SHADOWMEMORY_H_
#define SHADOWMEMORY_H_

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>

/*
* initShadowMemory() -- Initializes the Shadow Memory.  This must be called before any other Shadow Memory functions are called.
* freeShadowMemory should also be called before any subsequent calls to this function to prevent memory leakage.
*/

void initShadowMemory ();

/*
* freeShadowMemory() -- Frees all memory allocated by the Shadow Memory.
*/

void freeShadowMemory ();

/*
* store(uint64_t addr, uint32_t sizeInBytes, uint8_t *val) --Stores the series of bytes (up to sizeInBytes) stored in val into the shadow 
* memory at address addr.  Assumes val is at least sizeInBytes long.
*/

void store (uint64_t addr, uint32_t sizeInBytes, uint8_t * val);

/* 
* load(uint64_t addr, uint32_t size, uint8_t *buf) -- Retrieves size bytes from the address at addr in shadow memory 
* and stores them into the buffer at buf. Assumes buf is at least size long.
*/

void load (uint64_t addr, uint32_t sizeInBytes, uint8_t * buf);

/*
* chkDefaultVals() -- Checks to ensure the default map values are consistent.
*/

void chkDefaultVals ();

/*
* isDefault() -- Checks to see which chunks of memory have been allocated (and hence are no longer the default map values).
*/

void isDefault ();

/*
* reset() -- Resets the state of the shadow memory to be completely clean (that is, just like after a call to initShadowMemory()). 
*/

void reset ();

#endif /* SHADOWMEMORY_H_ */
