/*
* File: ShadowRegisters.h
* Author: David Raicher
* Description: Provides the shadow register backend for the trace reader.  Allows for a completely arbitrary amount
* of threads to be represented.  Each thread has a full set of registers as defined in LynxReg.  So the amount and size
* of registers should dynamically update with LynxReg.  The registers can be read or written at will.
*/

#ifndef _SHADOWREGISTERS_H_
#define _SHADOWREGISTERS_H_

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include "../shared/LynxReg.h"

/*
* initShadowRegisters(uint32_t numThreadsInit) -- Initalizes the shadow registers with the designated amount of threads.  This
* must be called before any other functions are called.
*/

void initShadowRegisters (uint32_t numThreadsInit);

/*
* freeShadowRegisters(void) -- Frees all memory allocated by the shadow registers.
*/

void freeShadowRegisters (void);

/*
* setThreadVal(LynxReg, uint32_t thread, uint8_t *val) -- Sets the value in the register represented by reg on the thread represented by
* thread to the sequence of bytes in val.  Assumes val is at least the length of the given register.
*/

void setThreadVal (LynxReg reg, uint32_t thread, uint8_t * val);

/*
* getRegisterVal(LynxReg reg, uint32_t thread) -- Returns a pointer to the beginning of the shadow register that represents reg in the
* thread number represented by thread.  If thread is larger than the amount of threads initialized in ShadowMemory, NULL is returned.
*/

const uint8_t *getThreadVal (LynxReg reg, uint32_t thread);

/*
* setFlagVal(uint32_t thread, uint64_t val) -- Sets the rflags/eflags on the given thread to the given val.
*/

//void setFlagVal (uint32_t thread, uint64_t val);

/*
* getFlagVal(uint32_t thread) -- Returns the value of the rflags/eflags on the given thread.
*/

//uint32_t getFlagVal (uint32_t thread);

#endif /* _SHADOWREGISTERS_H_ */
